@@ -1,5 +1,5 @@
 /*
-Copyright 2018 The Knative Authors.
+Copyright 2018 The Knative Authors
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -18,18 +18,16 @@ package revision
 
 import (
 	"context"
-	"errors"
-	"fmt"
 	"strings"
+	"time"
 
 	"github.com/google/go-containerregistry/pkg/authn/k8schain"
-	"golang.org/x/sync/errgroup"
 
 	corev1 "k8s.io/api/core/v1"
+	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/apimachinery/pkg/util/sets"
 	"k8s.io/client-go/kubernetes"
 	appsv1listers "k8s.io/client-go/listers/apps/v1"
-	corev1listers "k8s.io/client-go/listers/core/v1"
 	cachingclientset "knative.dev/caching/pkg/client/clientset/versioned"
 	clientset "knative.dev/serving/pkg/client/clientset/versioned"
 	revisionreconciler "knative.dev/serving/pkg/client/injection/reconciler/serving/v1/revision"
@@ -44,7 +42,8 @@ import (
 )
 
 type resolver interface {
-	Resolve(string, k8schain.Options, sets.String) (string, error)
+	Resolve(*v1.Revision, k8schain.Options, sets.String, time.Duration) ([]v1.ContainerStatus, error)
+	Clear(types.NamespacedName)
 }
 
 // Reconciler implements controller.Reconciler for Revision resources.
@@ -57,7 +56,6 @@ type Reconciler struct {
 	podAutoscalerLister palisters.PodAutoscalerLister
 	imageLister         cachinglisters.ImageLister
 	deploymentLister    appsv1listers.DeploymentLister
-	serviceLister       corev1listers.ServiceLister
 
 	resolver resolver
 }
@@ -65,16 +63,16 @@ type Reconciler struct {
 // Check that our Reconciler implements revisionreconciler.Interface
 var _ revisionreconciler.Interface = (*Reconciler)(nil)
 
-func (c *Reconciler) reconcileDigest(ctx context.Context, rev *v1.Revision) error {
+func (c *Reconciler) reconcileDigest(ctx context.Context, rev *v1.Revision) (bool, error) {
 	if rev.Status.ContainerStatuses == nil {
-		rev.Status.ContainerStatuses = make([]v1.ContainerStatuses, 0, len(rev.Spec.Containers))
+		rev.Status.ContainerStatuses = make([]v1.ContainerStatus, 0, len(rev.Spec.Containers))
 	}
 
 	if rev.Status.DeprecatedImageDigest != "" {
 		// Default old revisions to have ContainerStatuses filled in.
 		// This path should only be taken by "old" revisions that have exactly one container.
 		if len(rev.Status.ContainerStatuses) == 0 {
-			rev.Status.ContainerStatuses = append(rev.Status.ContainerStatuses, v1.ContainerStatuses{
+			rev.Status.ContainerStatuses = append(rev.Status.ContainerStatuses, v1.ContainerStatus{
 				Name:        rev.Spec.Containers[0].Name,
 				ImageDigest: rev.Status.DeprecatedImageDigest,
 			})
@@ -83,7 +81,8 @@ func (c *Reconciler) reconcileDigest(ctx context.Context, rev *v1.Revision) erro
 
 	// The image digest has already been resolved.
 	if len(rev.Status.ContainerStatuses) == len(rev.Spec.Containers) {
-		return nil
+		c.resolver.Clear(types.NamespacedName{Namespace: rev.Namespace, Name: rev.Name})
+		return true, nil
 	}
 
 	imagePullSecrets := make([]string, 0, len(rev.Spec.ImagePullSecrets))
@@ -97,70 +96,55 @@ func (c *Reconciler) reconcileDigest(ctx context.Context, rev *v1.Revision) erro
 		ImagePullSecrets:   imagePullSecrets,
 	}
 
-	var digestGrp errgroup.Group
-	containerStatuses := make([]v1.ContainerStatuses, len(rev.Spec.Containers))
-	for i, container := range rev.Spec.Containers {
-		container := container // Standard Go concurrency pattern.
-		i := i
-		digestGrp.Go(func() error {
-			digest, err := c.resolver.Resolve(container.Image,
-				opt, cfgs.Deployment.RegistriesSkippingTagResolving)
-			if err != nil {
-				return errors.New(v1.RevisionContainerMissingMessage(container.Image, fmt.Sprintf("failed to resolve image to digest: %v", err)))
-			}
-			if len(rev.Spec.Containers) == 1 || len(container.Ports) != 0 {
-				rev.Status.DeprecatedImageDigest = digest
-			}
-			containerStatuses[i] = v1.ContainerStatuses{
-				Name:        container.Name,
-				ImageDigest: digest,
-			}
-			return nil
-		})
-	}
-	if err := digestGrp.Wait(); err != nil {
+	statuses, err := c.resolver.Resolve(rev, opt, cfgs.Deployment.RegistriesSkippingTagResolving, cfgs.Deployment.DigestResolutionTimeout)
+	if err != nil {
+		// Clear the resolver so we can retry the digest resolution rather than
+		// being stuck with this error.
+		c.resolver.Clear(types.NamespacedName{Namespace: rev.Namespace, Name: rev.Name})
 		rev.Status.MarkContainerHealthyFalse(v1.ReasonContainerMissing, err.Error())
-		return err
+		return true, err
 	}
-	rev.Status.ContainerStatuses = containerStatuses
-	return nil
+	if len(statuses) > 0 {
+		rev.Status.ContainerStatuses = statuses
+
+		// For backwards-compatibility we need to continue to set the DeprecatedImageDigest field.
+		for i := range rev.Spec.Containers {
+			if len(rev.Spec.Containers) == 1 || len(rev.Spec.Containers[i].Ports) != 0 {
+				rev.Status.DeprecatedImageDigest = statuses[i].ImageDigest
+			}
+		}
+
+		return true, nil
+	}
+
+	// No digest yet, wait for re-enqueue when resolution is done.
+	return false, nil
 }
 
 func (c *Reconciler) ReconcileKind(ctx context.Context, rev *v1.Revision) pkgreconciler.Event {
-	readyBeforeReconcile := rev.Status.IsReady()
-
-	// We may be reading a version of the object that was stored at an older version
-	// and may not have had all of the assumed defaults specified.  This won't result
-	// in this getting written back to the API Server, but lets downstream logic make
-	// assumptions about defaulting.
-	rev.SetDefaults(ctx)
-	rev.Status.InitializeConditions()
+	readyBeforeReconcile := rev.IsReady()
 	c.updateRevisionLoggingURL(ctx, rev)
 
-	phases := []struct {
-		name string
-		f    func(context.Context, *v1.Revision) error
-	}{{
-		name: "image digest",
-		f:    c.reconcileDigest,
-	}, {
-		name: "user deployment",
-		f:    c.reconcileDeployment,
-	}, {
-		name: "image cache",
-		f:    c.reconcileImageCache,
-	}, {
-		name: "PA",
-		f:    c.reconcilePA,
-	}}
+	reconciled, err := c.reconcileDigest(ctx, rev)
+	if err != nil {
+		return err
+	}
+	if !reconciled {
+		// Digest not resolved yet, wait for background resolution to re-enqueue the revision.
+		rev.Status.MarkResourcesAvailableUnknown(v1.ReasonResolvingDigests, "")
+		return nil
+	}
 
-	for _, phase := range phases {
-		if err := phase.f(ctx, rev); err != nil {
+	for _, phase := range []func(context.Context, *v1.Revision) error{
+		c.reconcileDeployment,
+		c.reconcileImageCache,
+		c.reconcilePA,
+	} {
+		if err := phase(ctx, rev); err != nil {
 			return err
 		}
 	}
-
-	readyAfterReconcile := rev.Status.IsReady()
+	readyAfterReconcile := rev.Status.GetCondition(v1.RevisionConditionReady).IsTrue()
 	if !readyBeforeReconcile && readyAfterReconcile {
 		logging.FromContext(ctx).Info("Revision became ready")
 		controller.GetEventRecorder(ctx).Event(
@@ -168,7 +152,6 @@ func (c *Reconciler) ReconcileKind(ctx context.Context, rev *v1.Revision) pkgrec
 			"Revision becomes ready upon all resources being ready")
 	}
 
-	rev.Status.ObservedGeneration = rev.Generation
 	return nil
 }
 
@@ -178,9 +161,7 @@ func (c *Reconciler) updateRevisionLoggingURL(ctx context.Context, rev *v1.Revis
 		return
 	}
 
-	uid := string(rev.UID)
-
-	rev.Status.LogURL = strings.Replace(
+	rev.Status.LogURL = strings.ReplaceAll(
 		config.Observability.LoggingURLTemplate,
-		"${REVISION_UID}", uid, -1)
+		"${REVISION_UID}", string(rev.UID))
 }
